from django.conf import (
    settings,
)

from educommon.utils.seqtools import (
    make_chunks,
)

from edu_rdm_integration.adapters.runners import (
    WebEduRunner,
)


class BaseCollectingDataRunner(WebEduRunner):
    """
    Базовый класс ранеров функций сбора данных для интеграции с "Региональная витрина данных".

    Поддерживается режим принудительного запуска функций без постановки в очередь на исполнение.
    """

    def _get_runnable_objects(self, logs, *args, **kwargs):
        """
        Возвращает генератор запускаемых объектов.
        """
        if not self.forced_run:
            yield from super()._get_runnable_objects(raw_logs=logs, *args, **kwargs)
            return

        raw_logs_chunks = make_chunks(
            iterable=logs,
            size=settings.RDM_COLLECT_CHUNK_SIZE,
        )

        for chunk_index, raw_logs_chunk in enumerate(raw_logs_chunks, start=1):
            raw_logs = list(raw_logs_chunk)
            for runnable_class in self._prepare_runnable_classes():
                runnable = runnable_class(
                    raw_logs=raw_logs,
                    chunk_index=chunk_index,
                    *args,
                    **kwargs,
                )

                if isinstance(runnable, tuple):
                    yield from runnable
                else:
                    yield runnable



from datetime import (
    date,
    datetime,
)
from typing import (
    Any,
    List,
    Set,
    Tuple,
    Type,
)

from django.conf import (
    settings,
)

from educommon.utils.conversion import (
    str_without_control_chars,
)
from educommon.utils.crypto import (
    HashData,
)

from edu_rdm_integration.adapters.helpers import (
    WebEduFunctionHelper,
    WebEduRunnerHelper,
)
from edu_rdm_integration.consts import (
    DATE_FORMAT,
    EXPORT_DATETIME_FORMAT,
    HASH_ALGORITHM,
)
from edu_rdm_integration.export_data.base.caches import (
    BaseExportDataFunctionCacheStorage,
    BaseExportDataRunnerCacheStorage,
)


class BaseExportDataRunnerHelper(WebEduRunnerHelper):
    """
    Базовый класс помощников ранеров функций выгрузки данных для интеграции с "Региональная витрина данных".
    """

    def _prepare_cache_class(self) -> Type[BaseExportDataRunnerCacheStorage]:
        """
        Возвращает класс кеша помощника ранера.
        """
        return BaseExportDataRunnerCacheStorage


class BaseExportDataFunctionHelper(WebEduFunctionHelper):
    """
    Базовый класс помощников функций выгрузки данных для интеграции с "Региональная витрина данных".
    """

    cryptographer = HashData(hash_algorithm=HASH_ALGORITHM)

    @classmethod
    def _prepare_str_field(cls, field_value: str, *, required: bool) -> str:
        # Очистка строковых полей от управляющих символов
        return str_without_control_chars(field_value)

    @classmethod
    def _prepare_datetime_field(cls, field_value: datetime, *, required: bool) -> str:
        # Дату/время передаём в формате: YYYY-MM-DD hh:mm:ss
        return field_value.strftime(EXPORT_DATETIME_FORMAT)

    @classmethod
    def _prepare_date_field(cls, field_value: date, *, required: bool) -> str:
        return field_value.strftime(DATE_FORMAT)

    @classmethod
    def _prepare_common_field(cls, field_value: Any, *, required: bool) -> str:
        return str(field_value if field_value is not None else '')

    @classmethod
    def _surround_with_quotes(cls, field_value: str, *, required: bool) -> str:
        has_value = not (field_value is None or field_value == '')
        if not required and not has_value:
            return ''

        return f'"{field_value}"'

    @classmethod
    def prepare_record(
        cls,
        entity_instance,
        ordered_fields: Tuple[str, ...],
        primary_key_fields: Set[str],
        foreign_key_fields: Set[str],
        required_fields: Set[str],
        hashable_fields: Set[str],
        ignore_prefix_fields: Set[str],
    ) -> List[str]:
        """
        Формирование списка строковых значений полей.
        """
        field_values = []
        key_fields = primary_key_fields.union(foreign_key_fields)
        add_prefix_fields = key_fields.difference(ignore_prefix_fields)

        for field in ordered_fields:
            required = field in required_fields
            field_value = getattr(entity_instance, field)

            if isinstance(field_value, str):
                field_value = cls._prepare_str_field(field_value, required=required)
            elif isinstance(field_value, datetime):
                field_value = cls._prepare_datetime_field(field_value, required=required)
            elif isinstance(field_value, date):
                field_value = cls._prepare_date_field(field_value, required=required)
            else:
                field_value = cls._prepare_common_field(field_value, required=required)

            if field_value and field in add_prefix_fields:
                field_value = f'{settings.RDM_EXPORT_ENTITY_ID_PREFIX}-{field_value}'

            if field_value and field in hashable_fields:
                field_value = cls.cryptographer.get_hash(field_value)

            # Экранирование двойных кавычек
            field_value = field_value.replace('"', '""')

            field_value = cls._surround_with_quotes(field_value, required=required)

            field_values.append(field_value)

        return field_values

    def _prepare_cache_class(self) -> Type[BaseExportDataFunctionCacheStorage]:
        """
        Возвращает класс кеша помощника функции.
        """
        return BaseExportDataFunctionCacheStorage



import uuid
from datetime import (
    date,
    datetime,
    time,
)
from typing import (
    Iterable,
    List,
)

from django.db import (
    connection,
)

from m3_db_utils.models import (
    ModelEnumValue,
)

from edu_rdm_integration import (
    consts,
)
from edu_rdm_integration.models import (
    RegionalDataMartEntityEnum,
)


class BaseFirstExportEntitiesDataCommandsGenerator:
    """Класс, который генерирует список данных для формирования команд для экспорта данных РВД."""

    def __init__(
        self,
        entities: Iterable[str],
        period_started_at=datetime.combine(date.today(), time.min),
        period_ended_at=datetime.combine(date.today(), time.min),
        batch_size=consts.BATCH_SIZE,
        **kwargs,
    ):
        """Инициализация."""
        # Если сущности не указаны, берется значение по умолчанию - все сущности:
        entities = entities if entities else RegionalDataMartEntityEnum.get_enum_data().keys()
        self.entities: List[ModelEnumValue] = [
            RegionalDataMartEntityEnum.get_model_enum_value(entity) for entity in entities
        ]

        self.period_started_at = period_started_at
        self.period_ended_at = period_ended_at

        self.batch_size = batch_size

        self.generation_id = uuid.uuid4()

        # Правую дату нужно увеличивать на одну секунду, т.к. подрезались миллисекунды
        self.get_logs_periods_sql = """
            select min(modified),
                   max(modified) + interval '1 second',
                   row_batched
            from (
                select row_num,
                       ((row_num - 1) / {batch_size}) + 1 AS row_batched,
                       modified
                from (
                    select row_number() over (order by modified) as row_num,
                           modified
                    from (
                        select *
                        from (
                        {ordered_rows}
                        ) as union_rows
                        order by modified
                    ) as ordered
                ) as numbered
            ) as batched
            group by row_batched
            order by row_batched;
            """

        self.ordered_rows_query = """
            select distinct date_trunc('second', modified) as modified
            from {table_name}
            where modified between '{period_started_at}' and '{period_ended_at}'
            """

    def generate(self) -> List:
        """Генерирует список данных для формирования команд для экспорта данных РВД."""
        params_for_commands = []

        for entity in self.entities:
            ordered_rows_queries_sql = self.ordered_rows_query.format(
                table_name=entity.main_model_enum.model._meta.db_table,
                period_started_at=self.period_started_at.strftime(consts.DATE_FORMAT),
                period_ended_at=self.period_ended_at.strftime(consts.DATE_FORMAT),
            )

            temp_get_logs_periods_sql = self.get_logs_periods_sql.format(
                batch_size=self.batch_size,
                ordered_rows=ordered_rows_queries_sql,
            )

            with connection.cursor() as cursor:
                cursor.execute(temp_get_logs_periods_sql)
                rows = cursor.fetchall()

            for period_started_at, period_ended_at, _ in rows:
                params_for_commands.append(
                    {
                        'period_started_at': period_started_at,
                        'period_ended_at': period_ended_at,
                        'entity': entity.key,
                        'generation_id': self.generation_id,
                    },
                )

        return params_for_commands



from typing import (
    Tuple,
)

from edu_rdm_integration.adapters.helpers import (
    WebEduFunctionHelper,
    WebEduRunnerHelper,
)
from edu_rdm_integration.consts import (
    ALL_OPERATIONS,
    UPDATED_OPERATIONS,
)


class BaseCollectingDataRunnerHelper(WebEduRunnerHelper):
    """
    Базовый класс помощников ранеров функций сбора данных для интеграции с "Региональная витрина данных".
    """


class BaseCollectingDataFunctionHelper(WebEduFunctionHelper):
    """
    Базовый класс помощников функций сбора данных для интеграции с "Региональная витрина данных".
    """

    def get_filtered_operations(self, with_deleted: bool = False) -> Tuple[int]:
        """
        Возвращает кортеж отфильтрованных операций.
        """
        return ALL_OPERATIONS if with_deleted else UPDATED_OPERATIONS



from typing import (
    Type,
)

from edu_rdm_integration.collect_data.base.helpers import (
    BaseCollectingDataFunctionHelper,
    BaseCollectingDataRunnerHelper,
)
from edu_rdm_integration.collect_data.non_calculated.base.caches import (
    BaseCollectingExportedDataFunctionCacheStorage,
    BaseCollectingExportedDataRunnerCacheStorage,
)


class BaseCollectingExportedDataRunnerHelper(BaseCollectingDataRunnerHelper):
    """
    Базовый класс помощников ранеров функций сбора данных для интеграции с "Региональная витрина данных".
    """

    def _prepare_cache_class(self) -> Type[BaseCollectingExportedDataRunnerCacheStorage]:
        """
        Возвращает класс кеша помощника ранера.
        """
        return BaseCollectingExportedDataRunnerCacheStorage


class BaseCollectingExportedDataFunctionHelper(BaseCollectingDataFunctionHelper):
    """
    Базовый класс помощников функций сбора данных для интеграции с "Региональная витрина данных".
    """

    def _prepare_cache_class(self) -> Type[BaseCollectingExportedDataFunctionCacheStorage]:
        """
        Возвращает класс кеша помощника функции.
        """
        return BaseCollectingExportedDataFunctionCacheStorage



import heapq
import os
from concurrent.futures import (
    ThreadPoolExecutor,
)
from json import (
    JSONDecodeError,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)

from django.conf import (
    settings,
)
from django.db import (
    transaction,
)
from django.db.models import (
    CharField,
    OuterRef,
    QuerySet,
    Subquery,
)
from django.db.models.functions import (
    Cast,
    Least,
)
from uploader_client.adapters import (
    adapter,
)

from educommon import (
    logger,
)

from edu_rdm_integration.collect_and_export_data.models import (
    AbstractExportDataCommandProgress,
)
from edu_rdm_integration.enums import (
    FileUploadStatusEnum,
)
from edu_rdm_integration.export_data.base.requests import (
    RegionalDataMartStatusRequest,
)
from edu_rdm_integration.export_data.consts import (
    TOTAL_ATTACHMENTS_SIZE_KEY,
)
from edu_rdm_integration.models import (
    CollectingDataStageStatus,
    CollectingExportedDataStage,
    DataMartRequestStatus,
    ExportingDataStage,
    ExportingDataStageStatus,
    ExportingDataSubStageUploaderClientLog,
    UploadDataCommand,
    UploadStatusRequestLog,
)
from edu_rdm_integration.redis_cache import (
    AbstractCache,
)


if TYPE_CHECKING:
    from datetime import (
        datetime,
    )

    from uploader_client.models import (
        Entry,
    )

    from edu_rdm_integration.collect_data.non_calculated.base.managers import (
        BaseCollectingExportedDataRunnerManager,
    )
    from edu_rdm_integration.export_data.base.managers import (
        BaseExportDataRunnerManager,
    )


class UploadStatusHelper:
    """Хелпер проверки статуса загрузки данных в витрину."""

    def __init__(self, in_progress_uploads: QuerySet, cache: AbstractCache) -> None:
        self._in_progress_uploads = in_progress_uploads
        self.cache = cache

    def run(self, thread_count: int = 1) -> None:
        """Запускает проверки статусов."""
        if thread_count > 1:
            with ThreadPoolExecutor(max_workers=thread_count) as pool:
                pool.map(self._process_upload, self._in_progress_uploads)
        else:
            for upload in self._in_progress_uploads:
                self._process_upload(upload)

    @classmethod
    def send_upload_status_request(cls, request_id: str) -> Tuple[Optional[Dict[str, Any]], 'Entry']:
        """Формирует и отправляет запрос для получения статуса загрузки данных в витрину."""
        request = RegionalDataMartStatusRequest(
            request_id=request_id,
            method='GET',
            parameters={},
            headers={
                'Content-Type': 'application/json',
            },
        )

        result = adapter.send(request)

        response = None

        if result.error:
            logger.warning(
                f'Ошибка при получении статуса загрузки данных в витрину. Идентификатор загрузки: {request_id}. '
                f'Ошибка: {result.error}, запрос: {result.log.request}, ответ: {result.log.response}',
            )
        else:
            logger.info(
                f'Получен ответ со статусом {result.response.status_code} и содержимым {result.response.text}. '
                f'Идентификатор загрузки: {request_id}',
            )
            try:
                response = result.response.json()
            except JSONDecodeError:
                logger.error(
                    f'Не удалось получить данные из ответа запроса статуса загрузки данных в витрину. '
                    f'Идентификатор загрузки: {request_id}, ответ: {result.response.text}',
                )

        return response, result.log

    @classmethod
    def update_upload_status(
        cls,
        upload: ExportingDataSubStageUploaderClientLog,
        response: Optional[Dict[str, Any]],
        log_entry: 'Entry',
    ) -> None:
        """Обновляет статус загрузки данных в витрину."""
        request_status = None

        if isinstance(response, dict):
            request_status = DataMartRequestStatus.get_values_to_enum_data().get(response.get('code'))

            if not request_status:
                logger.error(
                    'Не удалось определить статус загрузки данных в витрину. Идентификатор загрузки: '
                    f'{upload.request_id}, данные ответа: {response}',
                )

        with transaction.atomic():
            UploadStatusRequestLog.objects.create(
                upload=upload,
                entry=log_entry,
                request_status_id=getattr(request_status, 'key', None),
            )

            if request_status in {DataMartRequestStatus.FAILED_PROCESSING, DataMartRequestStatus.REQUEST_ID_NOT_FOUND}:
                upload.file_upload_status = FileUploadStatusEnum.ERROR

            elif request_status == DataMartRequestStatus.SUCCESSFULLY_PROCESSED:
                upload.file_upload_status = FileUploadStatusEnum.FINISHED

            if upload.file_upload_status != FileUploadStatusEnum.IN_PROGRESS:
                upload.save()

    def _process_upload(self, upload: ExportingDataSubStageUploaderClientLog) -> None:
        """Обрабатывает запись загрузки данных в витрину."""
        response, log_entry = self.send_upload_status_request(upload.request_id)
        self.update_upload_status(upload, response, log_entry)
        # Обновим размер файлов в кеш (с блокировкой на время обновления)
        with self.cache.lock(f'{TOTAL_ATTACHMENTS_SIZE_KEY}:lock', timeout=300):
            queue_total_file_size = self.cache.get(TOTAL_ATTACHMENTS_SIZE_KEY) or 0
            if queue_total_file_size:
                queue_total_file_size -= upload.attachment.attachment_size
                if queue_total_file_size > 0:
                    self.cache.set(
                        TOTAL_ATTACHMENTS_SIZE_KEY,
                        queue_total_file_size,
                        timeout=settings.RDM_REDIS_CACHE_TIMEOUT_SECONDS
                    )


class Graph:
    """Граф связей между моделями.

    Предназначен для поиска кратчайшей связи между моделями и дальнейшего построения lookup`а
    до необходимого поля модели с последующим использованием его в фильтре.
    Вершинами графа выступают наименования моделей. Ребро содержит наименования модели связанной
    с другой моделью и наименования поля, через которое осуществляется связь.
    Вместо наименования поля может быть наименование обратной связи между моделями,
    в случае если данная связь является связью OneToOne.
    """

    def __init__(self):
        self.vertices: Dict[str, Dict[str, Optional[str]]] = {}
        """Словарь для хранения данных графа."""

    def add_vertex(self, vertex: str):
        """Добавление вершины."""
        if vertex not in self.vertices:
            self.vertices[vertex] = {}

    def add_edge(self, vertex1: str, vertex2: str, edge_name: str):
        """Добавление связи."""
        if vertex1 in self.vertices and vertex2 in self.vertices:
            self.vertices[vertex1][vertex2] = edge_name
            if vertex1 not in self.vertices[vertex2]:
                self.vertices[vertex2][vertex1] = None

    def remove_vertex(self, vertex: str):
        """Удаление вершины."""
        if vertex in self.vertices:
            del self.vertices[vertex]

            # Удаляем связанные с удаленной вершиной ребра
            for neighbour in self.vertices:
                if vertex in self.vertices[neighbour]:
                    self.vertices[neighbour].pop(vertex)

    def remove_edge(self, vertex1: str, vertex2: str):
        """Удаление связи."""
        if vertex1 in self.vertices and vertex2 in self.vertices:
            self.vertices[vertex1].pop(vertex2, None)
            self.vertices[vertex2].pop(vertex1, None)

    def get_vertices(self) -> List[str]:
        """Получение списка всех вершин."""
        return list(self.vertices)

    def get_edges(self) -> List[Tuple[str, str, Optional[str]]]:
        """Получение всех связей."""
        edges = []

        for vertex, neighbors in self.vertices.items():
            for neighboring_vertex, edge_name in neighbors.items():
                edge = (vertex, neighboring_vertex, edge_name)
                edges.append(edge)

        return edges

    def __contains__(self, vertex: str) -> bool:
        return vertex in self.vertices

    def __iter__(self):
        return iter(self.vertices)

    def __getitem__(self, vertex: str):
        return self.vertices.get(vertex, {})

    def get_edges_between_vertices(
        self, from_vertex: str, to_vertex: str, required_edge_name: bool = True
    ) -> List[str]:
        """Получение списка наименований ребер между вершинами."""
        if from_vertex not in self.vertices and to_vertex not in self.vertices:
            return []

        path = []
        edge_weight = 1

        # Инициализация расстояния между вершинами
        distances = {vertex: float('inf') for vertex in self}
        distances[from_vertex] = 0

        priority_queue = [(0, from_vertex)]

        while priority_queue:
            current_distance, current_vertex = heapq.heappop(priority_queue)

            # Если достигнута конечная вершина, заканчиваем цикл
            if current_vertex == to_vertex:
                break

            # Проверяем все смежные вершины и обновляем расстояния, если находим более короткий путь
            for neighbor, edge_name in self[current_vertex].items():
                distance = current_distance + edge_weight
                if distance < distances[neighbor]:
                    distances[neighbor] = distance
                    heapq.heappush(priority_queue, (distance, neighbor))

        # Восстанавливаем путь от конечной вершины к начальной
        current_vertex = to_vertex
        while current_vertex != from_vertex:
            previous_vertex = edge_name = None

            for neighbor in self[current_vertex]:
                if distances[current_vertex] == distances[neighbor] + 1:
                    for neighbor_vertex, edge_name in self[neighbor].items():
                        if neighbor_vertex == current_vertex:
                            break

                    previous_vertex = neighbor
                    break

            if required_edge_name and not edge_name:
                path = []
                break

            path.append(edge_name)
            current_vertex = previous_vertex

        # Инвертируем путь и возвращаем его
        path.reverse()

        return path


def save_command_log_link(
    command: Union[AbstractExportDataCommandProgress, UploadDataCommand],
    log_dir: str
) -> None:
    """Сохраняет ссылку на лог команды."""
    log_file = os.path.join(settings.MEDIA_ROOT, log_dir, f'{command.id}.log')
    if os.path.exists(log_file):
        command.logs_link = os.path.join(log_dir, f'{command.id}.log')
        command.save()


def get_collecting_managers_max_period_ended_dates(
    collecting_managers: Iterable['BaseCollectingExportedDataRunnerManager'],
) -> Dict[str, 'datetime']:
    """Возвращает дату и время завершения последнего успешного этапа сбора для менеджеров Функций сбора."""
    managers_last_period_ended = CollectingExportedDataStage.objects.filter(
        manager_id__in=[manager.uuid for manager in collecting_managers],
        id=Subquery(
            CollectingExportedDataStage.objects.filter(
                manager_id=OuterRef('manager_id'),
                status_id=CollectingDataStageStatus.FINISHED.key,
            ).order_by('-id').values('id')[:1]
        ),
    ).annotate(
        str_manager_id=Cast('manager_id', output_field=CharField()),
        last_period_ended_at=Least('logs_period_ended_at', 'started_at'),
    ).values_list(
        'str_manager_id', 'last_period_ended_at',
    )

    return {
        manager_id: last_period_ended_at
        for manager_id, last_period_ended_at in managers_last_period_ended
    }


def get_exporting_managers_max_period_ended_dates(
    exporting_managers: Iterable['BaseExportDataRunnerManager'],
) -> Dict[str, 'datetime']:
    """Возвращает дату и время последнего успешного этапа экспорта для менеджеров Функций экспорта."""
    managers_last_period_ended = ExportingDataStage.objects.filter(
        manager_id__in=[manager.uuid for manager in exporting_managers],
        id=Subquery(
            ExportingDataStage.objects.filter(
                manager_id=OuterRef('manager_id'),
                status_id=ExportingDataStageStatus.FINISHED.key,
            ).order_by('-id').values('id')[:1]
        ),
    ).annotate(
        str_manager_id=Cast('manager_id', output_field=CharField()),
        last_period_ended_at=Least('period_ended_at', 'started_at'),
    ).values_list(
        'str_manager_id', 'last_period_ended_at',
    )

    return {
        manager_id: last_period_ended_at
        for manager_id, last_period_ended_at in managers_last_period_ended
    }



# Generated by Django 3.1.14 on 2024-08-19 14:56

from django.db import (
    migrations,
)

from edu_rdm_integration.enums import (
    FileUploadStatusEnum,
)
from educommon.utils.seqtools import (
    make_chunks,
)


# Значение размера файла по умолчанию (если файл не найден)
ATTACHMENT_SIZES = 10_485_760
CHUNK_SIZE = 500


def set_attachment_size(apps, schema_editor):
    """Установка размера файла по умолчанию."""
    ExportingDataSubStageAttachment = apps.get_model(
        'edu_rdm_integration', 'ExportingDataSubStageAttachment'
    )  # noqa: N806
    ExportingDataSubStageUploaderClientLog = apps.get_model(
        'edu_rdm_integration', 'ExportingDataSubStageUploaderClientLog'
    )  # noqa: N806

    attachment_ids = ExportingDataSubStageUploaderClientLog.objects.filter(
        file_upload_status=FileUploadStatusEnum.IN_PROGRESS
    ).values_list('attachment_id', flat=True).iterator()

    sub_stage_with_deleted_attachments = []
    for chunked_attachment_ids in make_chunks(attachment_ids, CHUNK_SIZE):
        for sub_stage_attachment in ExportingDataSubStageAttachment.objects.filter(
            attachment_size__isnull=True,
            id__in=chunked_attachment_ids,
        ):
            if sub_stage_attachment.attachment.field.storage.exists(sub_stage_attachment.attachment.name):
                sub_stage_attachment.attachment_size = sub_stage_attachment.attachment.size
            else:
                sub_stage_attachment.attachment_size = ATTACHMENT_SIZES
            sub_stage_with_deleted_attachments.append(sub_stage_attachment)

        ExportingDataSubStageAttachment.objects.bulk_update(sub_stage_with_deleted_attachments, ['attachment_size'])


class Migration(migrations.Migration):

    dependencies = [
        ('edu_rdm_integration', '0012_exportingdatasubstageattachment_attachment_size'),
    ]

    operations = [
        migrations.RunPython(set_attachment_size, reverse_code=migrations.RunPython.noop)
    ]




from abc import (
    ABCMeta,
)
from typing import (
    List,
    Type,
    Union,
)

from django.core.files.base import (
    ContentFile,
)
from django.core.files.storage import (
    default_storage,
)
from transliterate import (
    slugify,
)

from educommon import (
    logger,
)
from educommon.integration_entities.enums import (
    EntityLogOperation,
)
from educommon.integration_entities.mixins import (
    EntitiesMixin,
)

from edu_rdm_integration.adapters.functions import (
    WebEduLazySavingPredefinedQueueGlobalHelperFunction,
)
from edu_rdm_integration.consts import (
    LOGS_DELIMITER,
)
from edu_rdm_integration.export_data.base.helpers import (
    BaseExportDataFunctionHelper,
)
from edu_rdm_integration.export_data.base.results import (
    BaseExportDataFunctionResult,
)
from edu_rdm_integration.export_data.base.validators import (
    BaseExportDataFunctionValidator,
)
from edu_rdm_integration.export_data.consts import (
    DELIMITER,
)
from edu_rdm_integration.models import (
    ExportingDataStage,
    ExportingDataSubStage,
    ExportingDataSubStageAttachment,
    ExportingDataSubStageEntity,
    ExportingDataSubStageStatus,
)
from edu_rdm_integration.utils import (
    get_exporting_data_stage_attachment_path,
)


class BaseExportDataFunction(
    EntitiesMixin,
    WebEduLazySavingPredefinedQueueGlobalHelperFunction,
    metaclass=ABCMeta,
):
    """
    Базовый класс функций выгрузки данных для интеграции с "Региональная витрина данных".
    """

    def __init__(self, *args, stage: ExportingDataStage, model_ids: List[Union[int, str]], **kwargs):
        super().__init__(*args, entities=self.entities, model_ids=model_ids, **kwargs)

        self._sub_stage = ExportingDataSubStage.objects.create(
            stage=stage,
            function_id=self.uuid,
        )
        ExportingDataSubStageEntity.objects.create(
            exporting_data_sub_stage=self._sub_stage,
            entity_id=self.first_entity.key
        )
        # Проставление подэтапа выгрузки
        self.entities[0].main_model_enum.model.objects.filter(pk__in=model_ids).update(
            exporting_sub_stage=self._sub_stage,
        )

        self._chunk_index = kwargs.get('chunk_index')

        logger.info(f'{LOGS_DELIMITER * 3}{repr(self._sub_stage)} created.')

        self._file_name = f'rdm_{self.first_entity.key.lower()}.csv'

        self._data = {
            EntityLogOperation.CREATE: [],
            EntityLogOperation.UPDATE: [],
            EntityLogOperation.DELETE: [],
        }

        self.has_data = False

    def _prepare_helper_class(self) -> Type[BaseExportDataFunctionHelper]:
        """
        Возвращает класс помощника функции.
        """
        return BaseExportDataFunctionHelper

    def _prepare_validator_class(self) -> Type[BaseExportDataFunctionValidator]:
        """
        Возвращает класс валидатора функции.
        """
        return BaseExportDataFunctionValidator

    def _prepare_result_class(self) -> Type[BaseExportDataFunctionResult]:
        """
        Возвращает класс результата функции.
        """
        return BaseExportDataFunctionResult

    def _before_prepare(self, *args, **kwargs):
        """
        Выполнение действий функций системы.
        """
        self._sub_stage.status_id = ExportingDataSubStageStatus.IN_PROGRESS.key
        self._sub_stage.save()

        logger.info(f'{LOGS_DELIMITER * 3}change status {repr(self._sub_stage)}')

    def _prepare_record(self, entity_instance) -> List[str]:
        """
        Формирование списка строковых значений полей.
        """
        ordered_fields = self.first_entity.entity.get_ordered_fields()
        primary_key_fields = set(self.first_entity.entity.get_primary_key_fields())
        foreign_key_fields = set(self.first_entity.entity.get_foreign_key_fields())
        required_fields = set(self.first_entity.entity.get_required_fields())
        hashable_fields = set(self.first_entity.entity.get_hashable_fields())
        ignore_prefix_fields = set(self.first_entity.entity.get_ignore_prefix_key_fields())

        field_values = self.helper.prepare_record(
            entity_instance=entity_instance,
            ordered_fields=ordered_fields,
            primary_key_fields=primary_key_fields,
            foreign_key_fields=foreign_key_fields,
            required_fields=required_fields,
            hashable_fields=hashable_fields,
            ignore_prefix_fields=ignore_prefix_fields,
        )

        return field_values

    def _prepare_data(self):
        """
        Преобразование собранных данных в удобный для выгрузки вид.
        """
        logger.info(f'{LOGS_DELIMITER * 3}{self.__class__.__name__} prepare data..')

        for entity_instance in self.helper.cache.entity_instances:
            self._data[entity_instance.operation].append(
                self._prepare_record(
                    entity_instance=entity_instance,
                )
            )

            entity_instance.exporting_sub_stage = self._sub_stage

        for operation in EntityLogOperation.values.keys():
            entities = self._data.get(operation)

            if entities:
                logger.info(
                    f'{LOGS_DELIMITER * 3}prepared {len(entities)} records with status '
                    f'{slugify(EntityLogOperation.values.get(operation))}..'
                )

    def _prepare_files(self):
        """
        Формирование файлов для дальнейшей выгрузки.
        """
        logger.info(f'{LOGS_DELIMITER * 3}{self.__class__.__name__} prepare files..')
        has_record = False

        for operation in EntityLogOperation.values.keys():

            records = self._data[operation]
            if records:
                title_record = (
                    f'{DELIMITER}'.join([
                        field.lower()
                        for field in self.first_entity.entity.get_ordered_fields()
                    ])
                )

                joined_records = (
                    '\n'.join([
                        title_record,
                        *[f'{DELIMITER}'.join(record) for record in records]]
                    )
                )

                sub_stage_attachment = ExportingDataSubStageAttachment(
                    exporting_data_sub_stage=self._sub_stage,
                    operation=operation,
                )

                file_path = get_exporting_data_stage_attachment_path(
                    instance=sub_stage_attachment,
                    filename=self._file_name,
                )

                sub_stage_attachment.attachment = default_storage.save(file_path, ContentFile(joined_records))
                sub_stage_attachment.attachment_size = sub_stage_attachment.attachment.size

                self.do_on_save(sub_stage_attachment)
                has_record |= True

        self.has_data = has_record

    def _prepare(self, *args, **kwargs):
        """
        Выполнение действий функции.
        """
        if self.result.has_not_errors:
            if self.helper.cache.entity_instances:
                self._prepare_data()
                self._prepare_files()
            else:
                logger.info(f'{LOGS_DELIMITER * 3} no data for preparing.')

    def before_run(self, *args, **kwargs):
        """
        Действия перед запуском.
        """
        super().before_run(*args, **kwargs)

        logger.info(
            '{delimiter}{force_run}run {runner_name}{log_chunks}..'.format(
                delimiter=LOGS_DELIMITER * 2,
                force_run='force ' if kwargs.get('is_force_run', False) else '',
                runner_name=self.__class__.__name__,
                log_chunks=f' with logs chunk {self._chunk_index}' if self._chunk_index else ''
            )
        )

    def run(self, *args, **kwargs):
        """
        Выполнение действий функции с дальнейшим сохранением объектов в базу при отсутствии ошибок.
        """
        super().run(*args, **kwargs)

        if self.result.has_not_errors:
            self._sub_stage.status_id = (
                ExportingDataSubStageStatus.FINISHED.key if not self.has_data
                else ExportingDataSubStageStatus.READY_FOR_EXPORT.key
            )
        else:
            self._sub_stage.status_id = ExportingDataSubStageStatus.FAILED.key

        self._sub_stage.save()

        logger.info(f'{LOGS_DELIMITER * 3}change status {repr(self._sub_stage)}')



from abc import (
    ABCMeta,
    abstractmethod,
)
from typing import (
    Tuple,
    Union,
)

from redis import (
    Redis,
    ResponseError,
)


def get_redis_version(connection: 'Redis') -> Tuple[int, int, int]:
    """Возвращает кортеж с версией сервера Redis."""
    try:
        version = getattr(connection, '__redis_server_version', None)
        if not version:
            version = tuple([int(n) for n in connection.info('server')['redis_version'].split('.')[:3]])
            setattr(connection, '__redis_server_version', version)
    except ResponseError:
        version = (0, 0, 0)

    return version


def as_text(v: Union[bytes, str]) -> str:
    """Конвертирует последовательность байт в строку."""
    if isinstance(v, bytes):
        return v.decode('utf-8')
    elif isinstance(v, str):
        return v
    else:
        raise ValueError('Неизвестный тип %r' % type(v))


class AbstractCache(metaclass=ABCMeta):
    """Абстрактный интерфейс для кеша отправки."""

    @abstractmethod
    def get(self, key, default=None, **kwargs):
        """Возвращает значение из кеша по ключу."""

    @abstractmethod
    def set(self, key, value, timeout=None, **kwargs):
        """Сохраняет значение в кеш по ключу."""

    @abstractmethod
    def lock(self, name, timeout=None, **kwargs):
        """Захватывает блокировку."""



import datetime
import decimal
import uuid
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Optional,
    Type,
)
from unittest import (
    mock,
)

from django.conf import (
    settings,
)
from django.db.models.fields.files import (
    FieldFile,
)
from django.test import (
    TestCase,
)
from django.utils import (
    timezone,
)

from educommon.audit_log.models import (
    AuditLog,
    Table,
)
from educommon.integration_entities.enums import (
    EntityLogOperation,
)
from educommon.utils.phone_number.phone_number import (
    PhoneNumber,
)
from function_tools.models import (
    Entity,
    EntityType,
)

from edu_rdm_integration.consts import (
    REGIONAL_DATA_MART_INTEGRATION_COLLECTING_DATA,
)
from edu_rdm_integration.models import (
    CollectingExportedDataStage,
    CollectingExportedDataSubStage,
)


if TYPE_CHECKING:
    from django.db.models import (
        Model,
    )

    from function_tools.managers import (
        RunnerManager,
    )


class BaseCollectingFunctionTestCase(TestCase):
    """Базовый класс тестирования Функции сбора."""

    databases = (settings.DEFAULT_DB_ALIAS, settings.SERVICE_DB_ALIAS)

    @property
    def manager(self) -> Type['RunnerManager']:
        """Менеджер раннера Функции сбора."""
        raise NotImplementedError

    @classmethod
    def tearDownClass(cls) -> None:
        """Вызывается один раз после запуска всех тестов класса."""
        cls.delete_auditlogs()

        super().tearDownClass()

    @staticmethod
    def delete_auditlogs() -> None:
        """Удаляет логи."""
        AuditLog.objects.all().delete()

    def setUp(self) -> None:
        """Подготавливает фикстуры."""
        self.now = timezone.now()

    def run_collecting_function(self) -> None:
        """Запускает Функцию сбора."""
        runner_manager = self.manager(
            logs_period_started_at=datetime.datetime.combine(self.now, datetime.time.min),
            logs_period_ended_at=datetime.datetime.combine(self.now, datetime.time.max),
        )
        runner_manager.run()

    @mock.patch('educommon.audit_log.models.AuditLog.ready_to_save')
    @mock.patch('educommon.audit_log.models.AuditLog.is_read_only')
    def create_auditlog(
        self,
        is_read_only_mock: mock.MagicMock,
        ready_to_save_mock: mock.MagicMock,
        *,
        instance: 'Model',
        changes: Optional[Dict[str, Any]] = None,
        operation: EntityLogOperation = EntityLogOperation.UPDATE,
    ) -> None:
        """Создает AuditLog."""
        is_read_only_mock.return_value = False
        ready_to_save_mock.return_value = True

        table = Table.objects.only('pk').get(name=instance._meta.db_table)

        timestamp = datetime.datetime.combine(self.now, timezone.now().time())

        if changes is None:
            changes = {'__stub': None}  # changes не может быть пустым
        else:
            changes['modified'] = timestamp.isoformat().replace('T', ' ')

        AuditLog.objects.create(
            user_id=1,
            user_type_id=1,
            ip='127.0.0.1',
            data=self._clean_data(instance),
            changes=changes,
            table=table,
            object_id=instance.pk,
            time=timestamp,
            operation=operation,
        )

    def create_sub_stage(self, *class_names: str) -> CollectingExportedDataSubStage:
        """Создает подэтап сбора данных."""
        function_tools_entities = dict(
            Entity.objects.filter(
                type__in=(EntityType.MANAGER.key, EntityType.FUNCTION.key),
                tags__contains=[REGIONAL_DATA_MART_INTEGRATION_COLLECTING_DATA],
                class_name__in=class_names,
            ).values_list('type', 'uuid'),
        )
        stage = CollectingExportedDataStage.objects.create(
            manager_id=function_tools_entities[EntityType.MANAGER.key],
        )

        return CollectingExportedDataSubStage.objects.create(
            function_id=function_tools_entities[EntityType.FUNCTION.key],
            stage=stage,
        )

    def _clean_data(self, instance: 'Model') -> Dict[str, Optional[str]]:
        """Подготавливает данные экземпляра модели для формирования AuditLog'a."""
        fields = {}

        for key, value in vars(instance).items():
            if key.startswith('_'):
                continue

            if isinstance(value, datetime.datetime):
                fields[key] = value.strftime('%Y-%m-%d %H:%M:%S.%f%z')
            elif isinstance(value, datetime.date):
                fields[key] = value.strftime('%Y-%m-%d')
            elif isinstance(value, datetime.time):
                fields[key] = value.strftime('%H:%M:%S')
            elif isinstance(value, bool):
                fields[key] = 't' if value else 'f'
            elif isinstance(value, (int, float, uuid.UUID, decimal.Decimal)):
                fields[key] = str(value)
            elif isinstance(value, FieldFile):
                fields[key] = value.name
            elif isinstance(value, list):
                fields[key] = f'{{{",".join(map(str, value))}}}'
            elif isinstance(value, PhoneNumber):
                fields[key] = value.cleaned
            else:
                fields[key] = value

        return fields
